@@ -14,12 +14,18 @@
 # ============================================================================
 """An implementation of a dm_env environment using dm_env_rpc."""
 
+import collections
+from typing import Any, Iterable, Mapping, Optional, Sequence, Tuple
 import dm_env
+import frozendict
 
+from dm_env_rpc.v1 import connection as dm_env_rpc_connection
 from dm_env_rpc.v1 import dm_env_flatten_utils
 from dm_env_rpc.v1 import dm_env_rpc_pb2
 from dm_env_rpc.v1 import dm_env_utils
+from dm_env_rpc.v1 import error
 from dm_env_rpc.v1 import spec_manager
+from dm_env_rpc.v1 import tensor_utils
 
 # Default observation names for common RL concepts.  By default the dm_env
 # wrapper will use these for reward and discount if available, but this behavior
@@ -32,13 +38,24 @@ DEFAULT_KEY_SEPARATOR = '.'
 
 
 class DmEnvAdaptor(dm_env.Environment):
-  """An implementation of dm_env using dm_env_rpc as the data protocol."""
+  """An implementation of dm_env using dm_env_rpc as the data protocol.
+
+  Users can also optionally provide a mapping of objects to DmEnvAdaptor
+  attributes. This is to accomodate user-created protocol extensions that
+  compliment the core protocol.
+
+  """
+
+  # Disable pytype attribute checking for dynamically created extension attrs.
+  _HAS_DYNAMIC_ATTRIBUTES = True
 
   def __init__(self,
-               connection,
-               specs,
-               requested_observations=None,
-               nested_tensors=True):
+               connection: dm_env_rpc_connection.Connection,
+               specs: dm_env_rpc_pb2.ActionObservationSpecs,
+               requested_observations: Optional[Sequence[str]] = None,
+               nested_tensors: bool = True,
+               extensions: Optional[Mapping[str,
+                                            Any]] = frozendict.frozendict()):
     """Initializes the environment with the provided dm_env_rpc connection.
 
     Args:
@@ -49,6 +66,8 @@ class DmEnvAdaptor(dm_env.Environment):
         environment when step is called. If None is specified then all
         observations will be requested.
       nested_tensors: Boolean to determine whether to flatten/unflatten tensors.
+      extensions: Optional mapping of extension instances to DmEnvAdaptor
+        attributes. Raises ValueError if attribute already exists.
     """
     self._dm_env_rpc_specs = specs
     self._action_specs = spec_manager.SpecManager(specs.actions)
@@ -89,6 +108,12 @@ class DmEnvAdaptor(dm_env.Environment):
     # Not strictly necessary but it makes the unit tests deterministic.
     self._requested_observation_uids.sort()
 
+    for extension_name, extension in extensions.items():
+      if hasattr(self, extension_name):
+        raise ValueError(
+            f'DmEnvAdaptor already has attribute "{extension_name}"!')
+      setattr(self, extension_name, extension)
+
   def reset(self):
     """Implements dm_env.Environment.reset."""
     response = self._connection.send(dm_env_rpc_pb2.ResetRequest())
@@ -228,3 +253,84 @@ class DmEnvAdaptor(dm_env.Environment):
     # Leaves the world if we were joined.  If not, this will be a no-op anyway.
     self._connection.send(dm_env_rpc_pb2.LeaveWorldRequest())
     self._connection = None
+
+
+def join_world(
+    connection: dm_env_rpc_connection.Connection,
+    world_name: str,
+    join_world_settings: Mapping[str, Any],
+    requested_observations: Optional[Iterable[str]] = None,
+    extensions: Optional[Mapping[str, Any]] = frozendict.frozendict()
+) -> DmEnvAdaptor:
+  """Helper function to join a world with the provided settings.
+
+  Args:
+    connection: An instance of Connection already connected to a dm_env_rpc
+      server.
+    world_name: Name of the world to join.
+    join_world_settings: Settings used to join the world. Values must be
+      packable into a Tensor message or already packed.
+    requested_observations: Optional set of requested observations.
+    extensions: Optional mapping of extension instances to DmEnvAdaptor
+      attributes.
+
+  Returns:
+    Instance of DmEnvAdaptor.
+  """
+
+  join_world_settings = {
+      key: (value if isinstance(value, dm_env_rpc_pb2.Tensor) else
+            tensor_utils.pack_tensor(value))
+      for key, value in join_world_settings.items()
+  }
+  try:
+    specs = connection.send(
+        dm_env_rpc_pb2.JoinWorldRequest(
+            world_name=world_name, settings=join_world_settings)).specs
+    return DmEnvAdaptor(
+        connection, specs, requested_observations, extensions=extensions)
+  except error.DmEnvRpcError:
+    connection.send(dm_env_rpc_pb2.LeaveWorldRequest())
+    raise
+
+
+def create_and_join_world(
+    connection: dm_env_rpc_connection.Connection,
+    create_world_settings: Mapping[str, Any],
+    join_world_settings: Mapping[str, Any],
+    requested_observations: Optional[Iterable[str]] = None,
+    extensions: Optional[Mapping[str, Any]] = frozendict.frozendict()
+) -> Tuple[DmEnvAdaptor, str]:
+  """Helper function to create and join a world with the provided settings.
+
+  Args:
+    connection: An instance of Connection already connected to a dm_env_rpc
+      server.
+    create_world_settings: Settings used to create the world. Values must be
+      packable into a Tensor proto or already packed.
+    join_world_settings: Settings used to join the world. Values must be
+      packable into a Tensor message.
+    requested_observations: Optional set of requested observations.
+    extensions: Optional mapping of extension instances to DmEnvAdaptor
+      attributes.
+
+  Returns:
+    Tuple of DmEnvAdaptor and the created world name.
+  """
+  create_world_settings = {
+      key: (value if isinstance(value, dm_env_rpc_pb2.Tensor) else
+            tensor_utils.pack_tensor(value))
+      for key, value in create_world_settings.items()
+  }
+  world_name = connection.send(
+      dm_env_rpc_pb2.CreateWorldRequest(
+          settings=create_world_settings)).world_name
+  try:
+    return_type = collections.namedtuple('DmEnvAndWorldName',
+                                         ['env', 'world_name'])
+    return return_type(
+        join_world(connection, world_name, join_world_settings,
+                   requested_observations, extensions), world_name)
+  except error.DmEnvRpcError:
+    connection.send(dm_env_rpc_pb2.DestroyWorldRequest(world_name=world_name))
+    raise
