@@ -15,6 +15,7 @@
 """Tests for dm_env_rpc helper functions."""
 
 from absl.testing import absltest
+from absl.testing import parameterized
 
 import numpy as np
 
@@ -38,7 +39,7 @@ class NpRangeInfoTests(absltest.TestCase):
       _ = tensor_spec_utils._np_range_info(np.str_).min
 
 
-class BoundsTests(absltest.TestCase):
+class BoundsTests(parameterized.TestCase):
 
   def test_unbounded_unsigned(self):
     tensor_spec = dm_env_rpc_pb2.TensorSpec()
@@ -231,22 +232,155 @@ class BoundsTests(absltest.TestCase):
     bounds = tensor_spec_utils.bounds(tensor_spec)
     self.assertEqual((-1, 1), bounds)
 
-  def test_max_larger_than_dtype_can_support_raises_error_legacy(self):
+  @parameterized.parameters([
+      dict(minimum=-500, maximum=0),
+      dict(minimum=0, maximum=500),
+  ])
+  def test_error_if_min_or_max_cannot_be_safely_cast_to_dtype(
+      self, minimum, maximum):
+    name = 'foo'
+    dtype = dm_env_rpc_pb2.DataType.INT8
     tensor_spec = dm_env_rpc_pb2.TensorSpec()
-    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT8
-    tensor_spec.max.int8 = 500
-    tensor_spec.name = 'foo'
-    with self.assertRaisesRegex(ValueError, 'foo.*-128, 500.*int8.*-128, 127'):
+    tensor_spec.dtype = dtype
+    tensor_spec.min.int8 = minimum
+    tensor_spec.max.int8 = maximum
+    tensor_spec.name = name
+    with self.assertRaisesWithLiteralMatch(
+        ValueError,
+        tensor_spec_utils._BOUNDS_CANNOT_BE_SAFELY_CAST_TO_DTYPE.format(
+            name=name,
+            minimum=minimum,
+            maximum=maximum,
+            dtype=dm_env_rpc_pb2.DataType.Name(dtype).lower(),
+        )):
       tensor_spec_utils.bounds(tensor_spec)
 
-  def test_min_larger_than_dtype_can_support_raises_error(self):
+  @parameterized.parameters([
+      dict(minimum=0., maximum=np.inf),
+      dict(minimum=-np.inf, maximum=0.),
+      dict(minimum=-np.inf, maximum=np.inf),
+  ])
+  def test_infinite_bounds_are_valid_for_floats(self, minimum, maximum):
     tensor_spec = dm_env_rpc_pb2.TensorSpec()
-    tensor_spec.dtype = dm_env_rpc_pb2.DataType.INT8
-    tensor_spec.min.int8 = -500
+    tensor_spec.dtype = dm_env_rpc_pb2.DataType.DOUBLE
+    tensor_spec.min.double = minimum
+    tensor_spec.max.double = maximum
     tensor_spec.name = 'foo'
-    with self.assertRaisesRegex(ValueError, 'foo.*-500, 127.*int8.*-128, 127'):
-      tensor_spec_utils.bounds(tensor_spec)
+    tensor_spec_utils.bounds(tensor_spec)
 
 
+class SetBoundsTests(parameterized.TestCase):
+
+  def test_set_scalar_bounds(self):
+    tensor_spec = dm_env_rpc_pb2.TensorSpec(
+        name='test', dtype=dm_env_rpc_pb2.DataType.INT32)
+    tensor_spec_utils.set_bounds(tensor_spec, minimum=1, maximum=2)
+
+    self.assertEqual([1], tensor_spec.min.int32s.array)
+    self.assertEqual([2], tensor_spec.max.int32s.array)
+
+  def test_set_multiple_bounds(self):
+    tensor_spec = dm_env_rpc_pb2.TensorSpec(
+        name='test', shape=(2,), dtype=dm_env_rpc_pb2.DataType.INT32)
+    tensor_spec_utils.set_bounds(tensor_spec, minimum=[1, 2], maximum=[3, 4])
+
+    self.assertEqual([1, 2], tensor_spec.min.int32s.array)
+    self.assertEqual([3, 4], tensor_spec.max.int32s.array)
+
+  def test_set_only_min(self):
+    tensor_spec = dm_env_rpc_pb2.TensorSpec(
+        name='test', dtype=dm_env_rpc_pb2.DataType.INT32)
+    tensor_spec_utils.set_bounds(tensor_spec, minimum=1, maximum=None)
+
+    self.assertEqual([1], tensor_spec.min.int32s.array)
+    self.assertIsNone(tensor_spec.max.WhichOneof('payload'))
+
+  def test_set_only_max(self):
+    tensor_spec = dm_env_rpc_pb2.TensorSpec(
+        name='test', dtype=dm_env_rpc_pb2.DataType.INT32)
+    tensor_spec_utils.set_bounds(tensor_spec, minimum=None, maximum=1)
+
+    self.assertIsNone(tensor_spec.min.WhichOneof('payload'))
+    self.assertEqual([1], tensor_spec.max.int32s.array)
+
+  def test_unset_min_and_max(self):
+    tensor_spec = dm_env_rpc_pb2.TensorSpec(
+        name='test', dtype=dm_env_rpc_pb2.DataType.INT32)
+    tensor_spec_utils.set_bounds(tensor_spec, minimum=1, maximum=2)
+    tensor_spec_utils.set_bounds(tensor_spec, minimum=None, maximum=None)
+
+    self.assertIsNone(tensor_spec.min.WhichOneof('payload'))
+    self.assertIsNone(tensor_spec.max.WhichOneof('payload'))
+
+  def test_cannot_set_nonnumeric_bounds(self):
+    tensor_spec = dm_env_rpc_pb2.TensorSpec(
+        name='test', dtype=dm_env_rpc_pb2.DataType.STRING)
+    with self.assertRaisesRegex(ValueError, 'non-numeric'):
+      tensor_spec_utils.set_bounds(tensor_spec, minimum=None, maximum=None)
+
+  def test_can_set_broadcastable_min_bounds(self):
+    tensor_spec = dm_env_rpc_pb2.TensorSpec(
+        name='test', shape=[2], dtype=dm_env_rpc_pb2.DataType.INT32)
+    tensor_spec_utils.set_bounds(tensor_spec, minimum=1, maximum=[2, 3])
+
+    self.assertEqual([1], tensor_spec.min.int32s.array)
+    self.assertEqual([2, 3], tensor_spec.max.int32s.array)
+
+  def test_cannot_set_multiple_min_bounds_on_variable_shape(self):
+    tensor_spec = dm_env_rpc_pb2.TensorSpec(
+        name='test', shape=[2, -1], dtype=dm_env_rpc_pb2.DataType.INT32)
+    with self.assertRaisesRegex(ValueError, 'incompatible'):
+      tensor_spec_utils.set_bounds(tensor_spec, minimum=[2, 3], maximum=4)
+
+  def test_cannot_set_dissimilar_shape_on_min_bounds(self):
+    tensor_spec = dm_env_rpc_pb2.TensorSpec(
+        name='test', shape=[2, 2], dtype=dm_env_rpc_pb2.DataType.INT32)
+    with self.assertRaisesRegex(ValueError, 'incompatible'):
+      tensor_spec_utils.set_bounds(tensor_spec, minimum=[1, 2, 3], maximum=4)
+
+  def test_can_set_broadcastable_max_bounds(self):
+    tensor_spec = dm_env_rpc_pb2.TensorSpec(
+        name='test', shape=[2], dtype=dm_env_rpc_pb2.DataType.INT32)
+    tensor_spec_utils.set_bounds(tensor_spec, minimum=[2, 3], maximum=4)
+
+    self.assertEqual([2, 3], tensor_spec.min.int32s.array)
+    self.assertEqual([4], tensor_spec.max.int32s.array)
+
+  def test_cannot_set_multiple_max_bounds_on_variable_shape(self):
+    tensor_spec = dm_env_rpc_pb2.TensorSpec(
+        name='test', shape=[2, -1], dtype=dm_env_rpc_pb2.DataType.INT32)
+    with self.assertRaisesRegex(ValueError, 'incompatible'):
+      tensor_spec_utils.set_bounds(tensor_spec, minimum=1, maximum=[2, 3])
+
+  def test_cannot_set_dissimilar_shape_on_max_bounds(self):
+    tensor_spec = dm_env_rpc_pb2.TensorSpec(
+        name='test', shape=[2, 2], dtype=dm_env_rpc_pb2.DataType.INT32)
+    with self.assertRaisesRegex(ValueError, 'incompatible'):
+      tensor_spec_utils.set_bounds(tensor_spec, minimum=0, maximum=[1, 2, 3])
+
+  def test_cannot_set_any_min_bounds_to_exceed_maximum(self):
+    tensor_spec = dm_env_rpc_pb2.TensorSpec(
+        name='test', shape=[2], dtype=dm_env_rpc_pb2.DataType.INT32)
+    with self.assertRaisesRegex(ValueError, 'larger than max'):
+      tensor_spec_utils.set_bounds(tensor_spec, minimum=[0, 4], maximum=[1, 1])
+
+  @parameterized.parameters([
+      dict(minimum=[-1000], maximum=[1]),
+      dict(minimum=[1], maximum=[1000]),
+  ])
+  def test_new_bounds_must_be_safely_castable_to_dtype(self, minimum, maximum):
+    name = 'test'
+    dtype = dm_env_rpc_pb2.DataType.INT8
+    tensor_spec = dm_env_rpc_pb2.TensorSpec(name=name, dtype=dtype)
+    with self.assertRaisesWithLiteralMatch(
+        ValueError,
+        tensor_spec_utils._BOUNDS_CANNOT_BE_SAFELY_CAST_TO_DTYPE.format(
+            name=name,
+            minimum=minimum,
+            maximum=maximum,
+            dtype=dm_env_rpc_pb2.DataType.Name(dtype))):
+      tensor_spec_utils.set_bounds(
+          tensor_spec, minimum=minimum, maximum=maximum)
+
 if __name__ == '__main__':
   absltest.main()
