@@ -317,7 +317,7 @@ class GetTensorTypeTests(absltest.TestCase):
       tensor_utils.get_tensor_type(mock_tensor)
 
 
-class GetTensorSpecTypeTests(absltest.TestCase):
+class DataTypeToNpTypeTests(absltest.TestCase):
 
   def test_float(self):
     self.assertEqual(
@@ -333,5 +333,46 @@ class GetTensorSpecTypeTests(absltest.TestCase):
       tensor_utils.data_type_to_np_type(30)
 
 
+class NpTypeToDataTypeTests(absltest.TestCase):
+
+  def test_float32(self):
+    self.assertEqual(
+        dm_env_rpc_pb2.DataType.FLOAT,
+        tensor_utils.np_type_to_data_type(np.float32))
+
+  def test_int32(self):
+    self.assertEqual(
+        dm_env_rpc_pb2.DataType.INT32,
+        tensor_utils.np_type_to_data_type(np.int32))
+
+  def test_dtype(self):
+    self.assertEqual(
+        dm_env_rpc_pb2.DataType.INT32,
+        tensor_utils.np_type_to_data_type(np.dtype(np.int32)))
+
+  def test_unknown_type(self):
+    with self.assertRaisesRegex(TypeError, 'dm_env_rpc DataType.*complex64'):
+      tensor_utils.np_type_to_data_type(np.complex64)
+
+
+class GetPackerTests(absltest.TestCase):
+
+  def test_cannot_get_packer_for_invalid_type(self):
+    with self.assertRaisesRegex(TypeError, 'complex64'):
+      tensor_utils.get_packer(np.complex64)
+
+  def test_can_pack(self):
+    packer = tensor_utils.get_packer(np.int32)
+    tensor = dm_env_rpc_pb2.Tensor()
+    packer.pack(tensor, np.asarray([1, 2, 3]))
+    self.assertEqual([1, 2, 3], tensor.int32s.array)
+
+  def test_can_unpack(self):
+    packer = tensor_utils.get_packer(np.int32)
+    tensor = dm_env_rpc_pb2.Tensor()
+    tensor.int32s.array[:] = [1, 2, 3]
+    np.testing.assert_array_equal([1, 2, 3], packer.unpack(tensor))
+
+
 if __name__ == '__main__':
   absltest.main()
